import numpy as np

_TREE_LEAF = -1
_TREE_UNDEFINED = -2

class TreeStruct(object):
    """ Basic Binary Tree Structure.
    
    
    Parameters
    ----------
        

    
    n_features : int
        Number of dimensions.
        
    Attributes
    ----------
    node_count : int
        Log of number of nodes.
        
    left_child : list
        Log of all left children nodes.
        
    right_child : list
        Log of all right children nodes.
        
    feature : array of int
        Log of splitting dimensions.
    
    threshold : array of float 
        Log of splitting points.
        
    n_node_samples : array of int
        Log of number of points in each node.
        
    node_range : list
        Log of boundary of nodes. 
    
    """
    def __init__(self, n_features):
        
        # Base tree statistic
        self.n_features = n_features
        self.node_count = 0
        # Base inner tree struct
        self.left_child = []
        self.right_child = []
        self.feature = []
        self.threshold = []
        self.n_node_samples = []
        self.leaf_ids = []
        self.leafnode_fun = {}  

        self.node_range = []
            
    def _node_append(self):
        """Add None to each logs as placeholders. 
        """
        self.left_child.append(None)
        self.right_child.append(None)
        self.feature.append(None)
        self.threshold.append(None)
        self.n_node_samples.append(None) 
        self.node_range.append(None)  
            
    def _add_node(self, parent, is_left, is_leaf, feature, threshold, n_node_samples, node_range=None):
        """Add a new node. 
        """
        self._node_append()
        node_id = self.node_count
        self.n_node_samples[node_id] = n_node_samples
        self.node_range[node_id] = node_range.copy()
            
        # record children status in parent nodes
        if parent != _TREE_UNDEFINED:
            if is_left:
                self.left_child[parent] = node_id
            else:
                self.right_child[parent] = node_id
        # record current node status
        if is_leaf:
            self.left_child[node_id] = _TREE_LEAF
            self.right_child[node_id] = _TREE_LEAF
            self.feature[node_id] = _TREE_UNDEFINED
            self.threshold[node_id] = _TREE_UNDEFINED
            self.leaf_ids.append(node_id)  
        else:
            # left_child and right_child will be set later
            self.feature[node_id] = feature
            self.threshold[node_id] = threshold
        self.node_count += 1
        return node_id
    
    def _node_info_to_ndarray(self):
        """Turn each logs into arrays. 
        """
        self.left_child = np.array(self.left_child, dtype=np.int32)
        self.right_child = np.array(self.right_child, dtype=np.int32)
        self.feature = np.array(self.feature, dtype=np.int32)
        self.threshold = np.array(self.threshold, dtype=np.float64)
        self.n_node_samples = np.array(self.n_node_samples, dtype=np.int32)
        self.leaf_ids = np.array(self.leaf_ids, dtype=np.int32)
        self.node_range = np.array(self.node_range, dtype=np.float64)
            
    def apply(self, X):
        """Get node ids.
        """
        return self._apply_dense(X)
    
    def _apply_dense(self, X):
        """Get node ids.
        """
        n = X.shape[0]
        result_nodeid = np.zeros(n, dtype=np.int32)
        for i in range(n):
            node_id = 0
            while self.left_child[node_id] != _TREE_LEAF:
                if X[i, self.feature[node_id]] < self.threshold[node_id]:
                    node_id = self.left_child[node_id]
                else:
                    node_id = self.right_child[node_id]
            result_nodeid[i] = node_id
        return  result_nodeid  

    def predict(self, X):
        """Predict for each x in X. 
        """
        node_affi = self.apply(X)
        y_predict_hat = np.zeros(X.shape[0])
        for leaf_id in self.leaf_ids:
            idx = node_affi == leaf_id
            
            y_predict_hat[idx] = self.leafnode_fun[leaf_id].predict(X[idx])
            
        return y_predict_hat


class RecursiveTreeBuilder(object):
    """ Recursively build a binary tree structure.
    
    Parameters
    ----------
    splitter : splitter object
        
    Estimator : estimator object
    
    epsilon : float
        Noise level.
    
    Y_range : float 
        Range of Y. 
        
    min_samples_split : int
        The minimum number of samples required to split an internal node.
    
    min_samples_leaf : int
        The minimum number of samples required in the subnodes to split an internal node.
    
    max_depth : int
        Maximum depth of the individual regression estimators.
        
    proportion_budget : float in [0,1]
        The propotion of privacy budgets spent on the numerator.

    min_ratio_pri_est : float
        If the privately estimated sample in the grid is less that n * min_ratio_pri_est, 
        use public estimation.
    
    """
    def __init__(self, splitter, 
                 Estimator, 
                 min_samples_split, 
                 min_samples_leaf,
                 max_depth,
                 epsilon,
                 Y_range,
                 proportion_budget,
                 min_ratio_pri_est,
                 ):
        
        # about splitter
        self.splitter = splitter
        # about estimator
        self.Estimator = Estimator
        # about recursive splits
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_depth = max_depth
        self.epsilon = epsilon
        self.Y_range = Y_range
        self.proportion_budget = proportion_budget
        self.min_ratio_pri_est = min_ratio_pri_est
        
        
    def build(self, tree, X, Y, X_range = None):
        """Grow the tree.
        """
        stack = []
        # prepare for stack [X, node_range, parent_status, left_node_status, depth]
        stack.append([X, Y, X_range, _TREE_UNDEFINED, _TREE_UNDEFINED, 0])
        while len(stack) != 0:
            dt_X, dt_Y, node_range, parent, is_left, depth = stack.pop()
            n_node_samples = dt_X.shape[0]
            # judge whether dt should be splitted or not
            if n_node_samples <= self.min_samples_split:
                is_leaf = True
            else:
                n_node_unique_samples = np.unique(np.hstack([dt_X,dt_Y.reshape(-1,1)]), axis=0).shape[0]
                # if too deep or contains too less samples, no split
                if depth >= self.max_depth or n_node_unique_samples <= self.min_samples_split:
                    is_leaf = True
                else:
                    rd_dim_vec, rd_split_vec = self.splitter(dt_X, node_range , dt_Y )
                    is_leaf = True
                    for idx_rd_dim in range(len(rd_dim_vec)):
                        rd_dim = rd_dim_vec[idx_rd_dim]
                        rd_split = rd_split_vec[idx_rd_dim]
                        if rd_split is None:
                            continue
                        elif (dt_X[:,rd_dim] >= rd_split).sum() < self.min_samples_leaf or (dt_X[:, rd_dim] < rd_split).sum() < self.min_samples_leaf:
                            continue
                        else:
                            is_leaf = False
                            break

            if not is_leaf:
                node_id = tree._add_node(parent, is_left, is_leaf, rd_dim, rd_split, n_node_samples, node_range)
            else:
                node_id = tree._add_node(parent, is_left, is_leaf, None, None, n_node_samples, node_range)
                tree.leafnode_fun[node_id] = self.Estimator(node_range, 
                                                            self.epsilon, 
                                                            self.Y_range, 
                                                            np.mean(dt_Y), 
                                                            self.proportion_budget,
                                                            self.min_ratio_pri_est,
                                                            )
                tree.leafnode_fun[node_id].fit()
            # begin branching if the node is not leaf
            if not is_leaf:
                # update node range status
                if node_range is not None:
                    node_range_right = node_range.copy()
                    node_range_left = node_range.copy()
                    node_range_right[0, rd_dim] = rd_split
                    node_range_left[1, rd_dim] = rd_split
                else:
                    node_range_right = node_range_left = None
                    
                # push right child on stack
                right_idx = dt_X[:,rd_dim] >= rd_split
                dt_X_right = dt_X[right_idx]
                dt_Y_right = dt_Y[right_idx]
                stack.append([dt_X_right, dt_Y_right , node_range_right, node_id, False, depth+1])
                
                # Push left child on stack
                left_idx = ~right_idx
                dt_X_left = dt_X[left_idx]
                dt_Y_left = dt_Y[left_idx]
                stack.append([dt_X_left, dt_Y_left, node_range_left, node_id, True, depth+1])
                
        tree._node_info_to_ndarray()
        
        
